/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */
package org.apache.groovy.ginq

import groovy.transform.CompileStatic
import org.apache.groovy.ginq.dsl.GinqAstBuilder
import org.apache.groovy.ginq.dsl.GinqAstVisitor
import org.apache.groovy.ginq.dsl.expression.GinqExpression
import org.apache.groovy.ginq.provider.collection.GinqAstWalker
import org.apache.groovy.lang.annotation.Incubating
import org.codehaus.groovy.ast.ClassHelper
import org.codehaus.groovy.ast.expr.ClosureExpression
import org.codehaus.groovy.ast.expr.ConstantExpression
import org.codehaus.groovy.ast.expr.Expression
import org.codehaus.groovy.ast.expr.MapEntryExpression
import org.codehaus.groovy.ast.expr.MapExpression
import org.codehaus.groovy.ast.stmt.Statement
import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.macro.runtime.Macro
import org.codehaus.groovy.macro.runtime.MacroContext

import static org.codehaus.groovy.ast.tools.GeneralUtils.asX
import static org.codehaus.groovy.ast.tools.GeneralUtils.mapX

/**
 * Declare GINQ macro methods
 *
 * @since 4.0.0
 */
@Incubating
@CompileStatic
class GinqGroovyMethods {
    /**
     * Transform GINQ code to target method invocation
     *
     * @param ctx the macro context
     * @param ginqClosureExpression hold the GINQ code
     * @return target method invocation
     * @since 4.0.0
     */
    @Macro
    static Expression GQ(final MacroContext ctx, final ClosureExpression ginqClosureExpression) {
        GQ(ctx, defaultConfiguration(), ginqClosureExpression)
    }

    /**
     * Represents the abbreviation of {@code GQ {...} as List}, which is very useful when used as list comprehension
     *
     * @param ctx the macro context
     * @param ginqClosureExpression hold the GINQ code
     * @return target method invocation
     * @since 4.0.0
     */
    @Macro
    static Expression GQL(final MacroContext ctx, final ClosureExpression ginqClosureExpression) {
        asX(ClassHelper.LIST_TYPE, GQ(ctx, ginqClosureExpression))
    }

    /**
     * Transform GINQ code to target method invocation
     *
     * @param ctx the macro context
     * @param ginqConfigurationMapExpression specify the configuration for GINQ, e.g. {@code astWalker}
     * @param ginqClosureExpression hold the GINQ code
     * @return target method invocation
     * @since 4.0.0
     */
    @Macro
    static Expression GQ(final MacroContext ctx, final MapExpression ginqConfigurationMapExpression, final ClosureExpression ginqClosureExpression) {
        return transformGinqCode(ctx.sourceUnit, ginqConfigurationMapExpression, ginqClosureExpression.code)
    }

    static Expression transformGinqCode(SourceUnit sourceUnit, MapExpression ginqConfigurationMapExpression, Statement code) {
        GinqAstBuilder ginqAstBuilder = new GinqAstBuilder(sourceUnit)
        code.visit(ginqAstBuilder)
        GinqExpression ginqExpression = ginqAstBuilder.getGinqExpression()

        Map<String, String> configuration = createConfiguration(ginqConfigurationMapExpression)
        Class<?> clazz = GinqGroovyMethods.class.classLoader.loadClass(configuration.get(CONF_AST_WALKER))
        GinqAstVisitor ginqAstWalker = (GinqAstVisitor) clazz.getDeclaredConstructor(SourceUnit.class).newInstance(sourceUnit)
        ginqAstWalker.setConfiguration(configuration)

        return (Expression) ginqAstWalker.visitGinqExpression(ginqExpression)
    }

    private static MapExpression defaultConfiguration() {
        return mapX(Arrays.asList(
                new MapEntryExpression(new ConstantExpression(CONF_AST_WALKER), new ConstantExpression(DEFAULT_AST_WALKER_CLASS_NAME))
        ))
    }

    private static Map<String, String> createConfiguration(MapExpression ginqConfigurationMapExpression) {
        Map<String, String> configuration = [:]
        for (MapEntryExpression mapEntryExpression : ginqConfigurationMapExpression.getMapEntryExpressions()) {
            configuration.put(mapEntryExpression.keyExpression.text, mapEntryExpression.valueExpression.text)
        }
        return configuration
    }

    private GinqGroovyMethods() {}

    private static final String CONF_AST_WALKER = 'astWalker'
    private static final String DEFAULT_AST_WALKER_CLASS_NAME = GinqAstWalker.class.name
}
